/**
 * Python Header
 * Must be included before any other headers.
 * Check the following for reference:
 * 1. https://docs.python.org/3/c-api/intro.html#include-files
 * 2. https://docs.python.org/3/c-api/arg.html#strings-and-buffers
 */
#define PY_SSIZE_T_CLEAN
#include <Python.h>
#include <structmember.h>
