#ifndef STREAMING_COMPRESSION_ZSTD_CONSTANTS_HPP
#define STREAMING_COMPRESSION_ZSTD_CONSTANTS_HPP

// C++ libraries
#include <cstddef>
#include <cstdint>

namespace streaming_compression { namespace zstd {
    constexpr int cDefaultCompressionLevel = 3;
} }

#endif // STREAMING_COMPRESSION_ZSTD_CONSTANTS_HPP
